#!/usr/bin/perl
#********************************************************************************
# IBM Spectrum Protect
# 
# name: tsmobjperf
#
# desc: Wrapper script to drive benchmark testing of an object storage system
#       from this host using the SPObjBench.jar Java tool. This script serves
#       as a convenience tool to drive a sequence of tests for modelling IBM 
#       Spectrum Protect-like workloads for assessing the suitability of the
#       object storage system for use with Spectrum Protect.
#       The tool is provided with the object storage type and credentials to
#       drive backup (PUT) and restore (GET) testing. A set of one or more
#       comma-delimited endpoints should be provided along with a test
#       bucket name to house test data. Note that this bucket must be
#       manually emptied/deleted afterwards. For "type=google", the endpoints
#       parameter can be left blank (https://storage.googleapis.com will be
#       used under the covers.
#       A min and max thread count is given along with a step. For example,
#       with min=10, max=100, step=10, the tool will test 10,20,30,...,100
#       thread write tests individually. If the special min of "1" is used,
#       step will be treated as starting from 0 (so test with 1,10,20,...100
#       threads in the previous example).
#       flist should be a comma-delimited list of 1 GB files. These should
#       all be the same size and should ideally be placed in a memory-mapped
#       file system to avoid any disk read bottlenecks. These will be uploaded
#       using multi-part upload to the bucket specified. The bucket should
#       already exist with appropriate permissions for the user credentials
#       to write to it. Data and/or the bucket will need to be deleted after
#       the tests complete. For each thread test iteration, 10 x thread_count GB
#       of data will be stored to the object storage system.
#       If restore tests are run as well, 256 KB range-reads are conducted
#       against the objects sent to object storage. By default, unless the
#       mode parameter is used, only write (ingest) tests are performed.
#       The tool uses the system-path Java to drive the underlying JAR tool.
#       Note that the user and pass parameter values should be enclosed in
#       double-quotes to avoid special character issues on the command-line.
#       An optional parameter can be used to select a specific java to use
#       to run the JAR tool. Otherwise, the java in the PATH is used.
#
# usage:  perl tsmobjperf.pl type=[s3|azure|google] 
#                            [endpoints=endpoint1,endpoint2,...] 
#                            user="[S3 Public Key]"
#                            pass="[S3 Private Key|Azure SAS Token]"
#                            gkey=[Google Cloud Key File]
#                            bucket=[S3 Bucket|Azure Container|Google Bucket]
#                            min=min_threads max=max_threads step=step
#                            flist=[file.1,file.2,...]
#                            [azblock=azure_block_size]
#                            [java=[/path/to/java]]
#                            [mode=write[|both]]
#
# requirements: This tool requires the following:
#                  - Run from a Perl interpreter
#                  - SPObjBench.jar is in the same local directory
#                  - Java 1.8+ JRE available in PATH or specified
#                  - Write permission to local directory for JAR output file
#
# tips:   - Use double-quotes around user and pass parameter values
#         - Remember to delete the content and/or the test bucket afterwards
#         - Be sure the right Java is chosen and has the HTTPS public key
#           of the endpoint imported, if necessary.
#
# Notice: This program is provided as a tool intended for use by IBM Internal, 
#         IBM Business Partners, and IBM Customers. This program is provided as is, 
#         without support, and without warranty of any kind expressed or implied.
#
# (C) Copyright International Business Machines Corp. 2018, 2020
#********************************************************************************
use Cwd;
use strict;


# Some utility variables
$|            = 1;
my $jar       = "-jar SPObjBench.jar";
my $teststamp = time();
my $dir       = getcwd();

# ----------------------------------------------------------------------
# Process command-line arguments ...
# ----------------------------------------------------------------------
my $type           = "";
my $endpointString = "";
my $user           = "";
my $pass           = "";
my $gkey           = "";
my $bucket         = "";
my $min            = 1;
my $max            = 1;
my $step           = 1;
my $flistString    = "";
my $azureBlockSize = 0;
my $java           = "java";
my $mode           = "write";
for my $arg ( @ARGV )
{
    if( $arg =~ m|([^=]+)=(.+)| )
    {
        my $cmd = $1;
        my $val = $2;
        if   ( lc($cmd) eq "type"      ) { $type           = lc($val); }
        elsif( lc($cmd) eq "endpoints" ) { $endpointString = $val;     }
        elsif( lc($cmd) eq "user"      ) { $user           = $val;     }
        elsif( lc($cmd) eq "pass"      ) { $pass           = $val;     }
        elsif( lc($cmd) eq "gkey"      ) { $gkey           = $val;     }
        elsif( lc($cmd) eq "bucket"    ) { $bucket         = $val;     }
        elsif( lc($cmd) eq "min"       ) { $min            = $val;     }
        elsif( lc($cmd) eq "max"       ) { $max            = $val;     }
        elsif( lc($cmd) eq "step"      ) { $step           = $val;     }
        elsif( lc($cmd) eq "flist"     ) { $flistString    = $val;     }
        elsif( lc($cmd) eq "azblock"   ) { $azureBlockSize  = $val;    }
        elsif( lc($cmd) eq "java"      ) { $java           = $val;     }
        elsif( lc($cmd) eq "mode"      ) { $mode           = lc($val); }
        else 
        {
            print "ERROR: Unrecognized parameter '$cmd'.\n";
            usage();
        }
    }
    else
    {
        print "ERROR: Unrecognized argument '$arg'.\n";
        usage();
    }
}

# Make sure SPObjBench.jar is in the local directory
if( ! -e $dir."/SPObjBench.jar" )
{
    print "ERROR: SPObjBench.jar not present in local directory.\n";
    usage();
}

# Error check some command-line input
if( ! ( $type eq "s3" ||  $type eq "azure" || $type eq "google" ) )
{
    print "ERROR: Parameter type must be 's3', 'azure', or 'google'.\n";
    usage();
}
if( $min <= 0 || $min > $max || $step < 1 )
{
    print "ERROR: Parameter min must be >0 and <= max and step must be >=1.\n";
    usage();
}
if( $type ne "google" && $endpointString eq "" )
{
    print "ERROR: One or more object storage endpoints must be specified when type != google.\n";
    usage();
}
if( $gkey ne "" && not -e $gkey )
{
    print "ERROR: Google key file '$gkey' does not exist.\n";
    usage();
}

# Check that each file provided in the flist is the same and 1 GB in size
my @files = split( /,/, $flistString );
my $size  = -1;
for my $file ( @files )
{
    if( ! -e $file ) { $size = -1; last; } # Error: file doesn't exist
    if( $size < 0 ) { $size = -s $file }
    elsif( $size != -s $file )
    {
        print "ERROR: Not all files in parameter flist match in size.\n";
        usage();
    }
}
if( $size == -1 )
{
    print "ERROR: Not all files exist in flist parameter listing.\n";
    usage();
}
if( $size != 1073741824 )
{
    print "ERROR: Files must all be 1 GB in size (1073741824 bytes).\n";
    usage();
}

my @endpointList = split(/,/, $endpointString);
my $endpointCount = scalar( @endpointList );

# Indicate the test that is to be performed
print "===================================================================\n";
print ": IBM Spectrum Protect object storage test\n";
print ":\n";
print ": Test Mode:\t\t$mode\n";
print ": Type:\t\t\t$type\n";
print ": Total Endpoints:\t$endpointCount\n";
print ": Min Threads:\t\t$min\n";
print ": Max Threads:\t\t$max\n";
print ": Thread Step:\t\t$step\n";
print ": File List:\t\t$flistString\n";
print ": Using java:\t\t$java\n";
print "===================================================================\n\n";

# Record JAR tool output for posterity
my $outfile = "tsmobjperf.$teststamp.out";
open( my $OUTFILE, ">$outfile" ) or die "Failed to open output file '$outfile' for JAR output.\n";
print "SPObjBench.jar output being captured to file:   $outfile\n\n";

# Report test results as we go
print "===================================================================\n";
print ": Test Results\n";
if( $mode eq "write" )
{
    print "Thread Count, Write Throughput (MB/s)\n";
}
else
{
    print "Thread Count, Write Throughput (MB/s), Read Throughput (MB/s)\n";
}


# Starting with min and incrementing by step up to max, test file upload and
# restore in Spectrum Protect-like fashion using the JAR tool
my $threads = $min;
while( $threads <= $max )
{
    # Build the command we will use here
    # Note that we limit testing to 10 objects per thread or 5 minutes, whichever comes first
    my $cmd = "";
    if( $type eq "google" )
    {
      $cmd .= "$java $jar -c $type -g $gkey -b $bucket -s $size -r $threads -o 10 -t 300";
    }
    else
    {
      $cmd .= "$java $jar -c $type -e $endpointString -u \"$user\" -p \"$pass\" -b $bucket -s $size -r $threads -o 10 -t 300";
    }
    if   ( $type eq "s3" || $type eq "google"      ) { $cmd .= " -K 134217728:100";   }
    elsif( $type eq "azure" && $azureBlockSize > 0 ) { $cmd .= " -P $azureBlockSize"; }
    $cmd .= " -S $flistString";

    # Check if write or write and read
    if( $mode eq "both" )
    {
        # Limit to 5-minute, random 256 KB range-read test
        $cmd .= " -R mixed:262144 -T 300";
    }

    # Capture standard error as well
    $cmd .= " 2>\&1";
    #print "$cmd\n";
    #exit();

    # Execute the command and capture/parse output
    my $tput = -1;
    my $rtput = -1;
    chomp( my $output = `$cmd` );
    my @lines = split( /\n/, $output );
    for my $line ( @lines )
    {
        if( $line =~ m|PUT performance in MBytes/s:\s+([0-9\.]+)| )
        {
            # A huge JAR tput of 999999.. indicates an issue
            $tput = $1;
            if( $tput =~ m|99999|) { $tput = -1; }
            if( $mode ne "both" )
            {
                last;
            }
        }
        elsif( $mode eq "both" && $line =~ m|GET performance in MBytes/s:\s+([0-9\.]+)| )
        {
            # A huge JAR tput of 999999.. indicates an issue
            $rtput = $1;
            if( $tput =~ m|99999|) { $rtput = -1; }
            last;
        }
        elsif( $line =~ m|Exception| )
        {
            # Error encounted with JAR
            last;
        }
    }

    # Capture output to file
    print $OUTFILE "Command: $cmd\n\n";
    print $OUTFILE "$output\n\n";

    # Error check output
    if( $tput == -1 )
    {
        # Notify user
        print "$threads, ERROR ENCOUNTERED\n";
        
        # Send output to file
        print "Error encountered with underlying SPObjBench.jar tool. See '$outfile' for more information.\n";
        last;
    }

    # Report on the thread count and throughput
    if( $mode eq "both" )
    {
        print "$threads, $tput, $rtput\n";
    }
    else
    {
        print "$threads, $tput\n";
    }

    # Increment thread count. Handle special '1' thread case
    if( $threads == 1 && $step > 1 ) { $threads += $step - 1; }
    else                             { $threads += $step;     }
}

# Done
print "===================================================================\n";
close( $OUTFILE );



# Utility sub-routine to print out usage and exit
sub usage
{
    print   "USAGE: tsmobjperf.pl type=[s3|azure|google] endpoints=[endpoint1,endpoint2,...]\n".
            "                     user=\"[S3 Public Key]\"\n".
            "                     pass=\"[S3 Private Key|Azure SAS Token]\"\n".
            "                     gkey=\"[Google Key File]\"\n".
            "                     bucket=[S3 Bucket|Azure Container|Google Bucket]\n".
            "                     min=min_threads max=max_threads step=step\n".
            "                     flist=[file.1,file.2,...]\n".
            "                     [azblock=azure_block_size]\n".
            "                     [java=[/path/to/java]]\n".
            "                     [mode=write[|both]]\n";
    exit(1);
}